/*
 * This module models the behavior of a multi-lift systems.
 * The system consists of NumberofLifts lifts and NumberofFloors floors.
 * Each lift can move a distance of one floor in 1s
 */


#include "liftsystem.h"
#include "global.h"


LiftSystem::LiftSystem(sc_module_name name, unsigned int lifts, unsigned int floors) : sc_module(name) {

	// intialization
	NumberofLifts = lifts;
	NumberofFloors = floors;

	// init the requests
	InternalRequests = std::vector< std::vector< int > >(lifts, std::vector< int >(floors, -1));
	ExternalRequestsUp = std::vector< int >(floors, -1);
	ExternalRequestsDown = std::vector< int >(floors, -1);

	// init the lifts - at beginning all lifts are at bottom floor and idle
	for (int i = 0; i < lifts; i++) {
		lift_status_t l;
		l.direction = 0;
		l.floor = 0;
		LiftStatus.push_back(l);
	}

	// passby
	passby = 0;

	// thread - get requests
	SC_THREAD(GetRequests);
	for (int i = 0; i < NUMBEROFUSERS; i++) {
		sensitive << reqs_p[i];
	}
	// notified when receving requests
	SC_METHOD(ReceiveRequests);
	for (int i = 0; i < NUMBEROFUSERS; i++) {
		sensitive << reqs_p[i];
	}
	dont_initialize();

	// thread - update lifts
	SC_THREAD(UpdateLifts);
}

void LiftSystem::ReceiveRequests() {

	receive_req_e.notify( SC_ZERO_TIME );
}


/*
 * control process of the controller
 */
void LiftSystem::GetRequests() {

	while (true) {
		wait(receive_req_e);

		int type = -1;
		int fromfloor = -1;
		int targetfloor = -1;
		int inlift = -1;
		int direction = 0;

		for (int i = 0; i < NUMBEROFUSERS; i++) {
			std::stringstream req;
			req << reqs_p[i].read();

			// get the request's element
			req >> type;
			req >> fromfloor;
			req >> targetfloor;
			req >> inlift;
			req >> direction;
			req.clear();

			// get internal request
			if (type == 0) {
				// assign it
				if (inlift != -1 && targetfloor != -1) {
					AssignInternalRequest(inlift, targetfloor);
				}
			}
			// get external request
			else if (type == 1) {
				// assign it
				if (fromfloor != -1 && direction != 0) {
					AssignExternalRequest(fromfloor, direction);
				}
			}
		}
	}
}

/*
 * Passby check
 */
void LiftSystem::PassbyCheck() {
	passby = 0;

	update_exreqs_up.lock();
	update_exreqs_down.lock();
	update_liftstatus.lock();
	// passby check
	for (int i = 0; i < NumberofFloors; i++) {
		for (int j = 0; j < NumberofLifts; j++) {
			if (ExternalRequestsUp[i] > -1) {
				if (LiftStatus[j].floor == i && LiftStatus[j].direction == 1 && ExternalRequestsUp[i] != j) {
					passby = 1;
					#ifdef DEBUG_MODEL
    				std::cout << "Passby external request up" << std::endl;
    				#endif
				} 
			}

			if (ExternalRequestsDown[i] > -1) {
				if (LiftStatus[j].floor == i && LiftStatus[j].direction == -1 && ExternalRequestsDown[i] != j) {
					passby = 1;
					#ifdef DEBUG_MODEL
    				std::cout << "Passby external request down" << std::endl;
    				#endif
				} 
			}
		}
	}
	update_liftstatus.unlock();
	update_exreqs_down.unlock();
	update_exreqs_up.unlock();
}

/*
 * Behaviors of lifts
 */
void LiftSystem::UpdateLifts() {

	while (true) {

		// update the lifts status		
		for (int i = 0; i < NumberofLifts; i++) {
			// update the status of lifts
			UpdateLiftStatus(i);
		}

		// each 1s lifts can move a distance of one floor
		wait(1, SC_SEC);
		// move the lifts
		for (int i = 0; i < NumberofLifts; i++) {
			MovingLift(i);
		}

		// passby check
		PassbyCheck();
	}
}

/*
 * Add an internal request in the lift "lift" to floor "level"
 */
void LiftSystem::AssignInternalRequest(int lift, int level) {

	// lock mutex
	update_intreqs.lock(); // wait until mutex is available

	#ifdef DEBUG_MODEL
    std::cout << "Add interal request (lift: " << lift << ", floor: " << level  << ")" << std::endl;
    #endif
    InternalRequests[lift][level] = 1;
	
	update_intreqs.unlock();
}


int LiftSystem::ComputeMinimumDistance(int level, int up) {
	int minimumDistance = INT_MAX;
	int chosenLift = -1;

	for (int i = 0; i < NumberofLifts; i++) {
		int distance = -1;
		
		update_liftstatus.lock(); // wait till mutex is available

		lift_status_t currlift = LiftStatus[i];

		// going up request
		if (up == 1) {
			// current lift is going up
			if (currlift.direction == 1) {
				if (currlift.floor  <= level)
					distance = level - currlift.floor;
				else 
					distance = (NumberofFloors - 1 - currlift.floor) + (NumberofFloors -1 - level);
			}
			// current lift is waiting
			else if (currlift.direction == 0) {
				if (currlift.floor  <= level)
					distance = level - currlift.floor;
				else 
					distance = currlift.floor - level;
			}
			// current lift is going down
			else {
					distance = currlift.floor + level;
			}
		}
		// going down request
		else {
			// current lift is going up
			if (currlift.direction == 1) {
				if (currlift.floor  <= level)
					distance = (NumberofFloors - 1 - currlift.floor) + (NumberofFloors -1 - level);
			}
			// current lift is waiting
			else if (currlift.direction == 0) {
				if (currlift.floor  <= level)
					distance = level - currlift.floor;
				else 
					distance = currlift.floor - level;
			}
			// current lift is going down
			else {
				if (currlift.floor  < level)
					distance = currlift.floor + level;
				else 
					distance = currlift.floor - level;
			}
		}

		// update the distance
		if (distance < minimumDistance) {
			chosenLift = i;
			minimumDistance = distance;
		}

		update_liftstatus.unlock();
	}

	return chosenLift;
}

/*
 * Add an external request at floor "level" with direction "up"
 */
void LiftSystem::AssignExternalRequest(int level, int up) {

	// find the lift with minimum distance from the floor "level"
	int chosenLift = ComputeMinimumDistance(level, up);
	
	// at ground floor, going down is not possible
	if (level == 0 && up == -1) {
		return;
	}

	// at top floor, going up is not possible

	if (level == (NumberofFloors - 1) && up == 1) {
		return;
	}

	if (up == 1) {
		update_exreqs_up.lock();
		// external request at floor "level" to go up is already requested
		if (ExternalRequestsUp[level] >= 0) {
			update_exreqs_up.unlock();
			return;
		}
		// add new external request to go up
		else {
			#ifdef DEBUG_MODEL
    		std::cout << "Add external request (floor: " << level << ", direction: " << up << ")" << std::endl;
    		#endif
			ExternalRequestsUp[level] = chosenLift;
			update_exreqs_up.unlock();
		}
	}

	if (up == -1) {
		update_exreqs_down.lock();
		// external request at floor "level" to go down is already requested
		if (ExternalRequestsDown[level] >= 0) {
			update_exreqs_down.unlock();
			return;
		}
		// add new request to go down
		else {
			#ifdef DEBUG_MODEL
    		std::cout << "Add external request (floor: " << level << ", direction: " << up << ")" << std::endl;
    		#endif
			ExternalRequestsDown[level] = chosenLift;
			update_exreqs_down.unlock();
		}
	}
}


int LiftSystem::GetFloortoMove_Int(int fromfloor, std::vector < int > targets) {
	
	int minidistance = INT_MAX;
	int chosenfloor = -1;
	int distance = -1;

	for (int i = 0; i < targets.size(); i++) {

		if (targets[i] >= fromfloor)
			distance = targets[i] - fromfloor;
		else
			distance = fromfloor - targets[i];

		// update the distance
		if (distance < minidistance) {
			minidistance = distance;
			chosenfloor = i; 
		}
	}

	return chosenfloor;
}

int LiftSystem::GetFloortoMove_Ext(int fromfloor, std::vector < int > targets, int direction) {

	int minidistance = INT_MAX;
	int chosenfloor = -1;
	int distance = -1;

	for (int i = 0; i < targets.size(); i++) {
	
		if (direction == 0) {
			if (targets[i] >= fromfloor)
				distance = targets[i] - fromfloor;
			else
				distance = fromfloor - targets[i];
		}
		else if (direction == 1) {
			if (targets[i] >= fromfloor)
				distance = targets[i] - fromfloor;
			else
				distance = (NumberofFloors - 1 - fromfloor) + (NumberofFloors -1 - targets[i]);
		}
		else {
			if (targets[i] <= fromfloor)
				distance = fromfloor - targets[i];
			else
				distance = fromfloor + targets[i];
		}

		// update the distance
		if (distance < minidistance) {
			minidistance = distance;
			chosenfloor = i; 
		}

	}

	return chosenfloor;
}


/* 
 * update the lift postions
 */
void LiftSystem::UpdateLiftStatus(int lift) {

	bool becomewaiting = true;

	// get the lift
	update_liftstatus.lock();

	lift_status_t l = LiftStatus[lift];
	// check exreq up
	update_exreqs_up.lock(); // wait until mutex is available

    if (ExternalRequestsUp[l.floor] == lift) {
    	#ifdef DEBUG_MODEL
    	std::cout << "Clear external (up) request for lift " << lift << " at floor " << l.floor << " direction " << l.direction << std::endl;
    	#endif
    	ExternalRequestsUp[l.floor] = -1;
    }
    update_exreqs_up.unlock(); // unlock mutex


    // check exreq down
	update_exreqs_down.lock(); // wait until mutex is available

    if (ExternalRequestsDown[l.floor] == lift) {
    	#ifdef DEBUG_MODEL
    	std::cout << "Clear external (down) request for lift " << lift << " at floor " << l.floor << " direction " << l.direction << std::endl;
    	#endif
    	ExternalRequestsDown[l.floor] = -1;
    }
    update_exreqs_down.unlock(); // unlock mutex

    // check intreq
	update_intreqs.lock(); // wait until mutex is available
   
    // the lift arrived
    if (InternalRequests[lift][l.floor] > -1) {
    	#ifdef DEBUG_MODEL
    	std::cout << "Clear internal request for lift " << lift << " at floor " << l.floor << " direction " << l.direction << std::endl;
    	#endif
    	InternalRequests[lift][l.floor] = -1;
    }

    // become waiting
    for (int i = 0; i < InternalRequests[lift].size(); i++) {
    	if (InternalRequests[lift][i] > -1) {
    		becomewaiting = false;
    	}
    }
    if (becomewaiting) {
    	l.direction = 0;
    }

    update_intreqs.unlock(); // unlock mutex
	
	update_liftstatus.unlock();
}


void LiftSystem::MovingLift(int lift) {

	std::vector < int > isUpdate_ExreqUp;
	std::vector < int > isUpdate_ExreqDown;
	std::vector < int > isUpdate_Intreq;
	int chosenfloor = -1;

	update_liftstatus.lock();

	// get the lift
	lift_status_t l = LiftStatus[lift];

	// check exreq up
	isUpdate_ExreqUp.clear();

	update_exreqs_up.lock(); // wait until mutex is available

    for (int i = 0; i < ExternalRequestsUp.size(); i++) {
    	if (ExternalRequestsUp[i] == lift) {
    		if (l.floor == i) {
    			ExternalRequestsUp[i] = -1;
    		}
    		else {
    			isUpdate_ExreqUp.push_back(i);
    		}
    	}
    }
    update_exreqs_up.unlock(); // unlock mutex

    // check exreq down
	isUpdate_ExreqDown.clear();

	update_exreqs_down.lock(); // wait until mutex is available

    for (int i = 0; i < ExternalRequestsDown.size(); i++) {
    	if (ExternalRequestsDown[i] == lift) {
    		if (l.floor == i) {
    			ExternalRequestsDown[i] = -1;
    		}
    		else {
    			isUpdate_ExreqDown.push_back(i);
    		}
    	}
    }
    update_exreqs_down.unlock(); // unlock mutex

    // check intreq
	isUpdate_Intreq.clear();

	update_intreqs.lock(); // wait until mutex is available
   
    for (int i = 0; i < InternalRequests[lift].size(); i++) {
    	if (InternalRequests[lift][i] > -1) {
    		if (l.floor == i) {
    			InternalRequests[lift][i] = -1;
    		}
    		else {
    			isUpdate_Intreq.push_back(i);
    		}
    	}
    }
    update_intreqs.unlock(); // unlock mutex
	
    // update moving based on requests
    if (isUpdate_Intreq.size() > 0) {
    	
    	chosenfloor = GetFloortoMove_Int(l.floor, isUpdate_Intreq);
    	if (chosenfloor != -1) {
    		if (l.floor == NumberofFloors - 1) {
				// go down
				LiftStatus[lift].direction = -1;
				LiftStatus[lift].floor = LiftStatus[lift].floor - 1;
			}

			else if (l.floor == 0) {
				// go up
				LiftStatus[lift].direction = 1;
				LiftStatus[lift].floor = LiftStatus[lift].floor + 1;
			}

			else if (l.floor > chosenfloor && l.floor > 0) {
				// go down
				LiftStatus[lift].direction = -1;
				LiftStatus[lift].floor = LiftStatus[lift].floor - 1;
			}
		
			else if (l.floor < chosenfloor && l.floor < NumberofFloors - 1) {
				// go up
				LiftStatus[lift].direction = 1;
				LiftStatus[lift].floor = LiftStatus[lift].floor + 1;
			}
		}
    }
    // there is no internal requests
    else {
    	// set the lift to be idle
    	LiftStatus[lift].direction = 0;

    	int chosenfloor_up = -1;
    	int chosenfloor_down = -1;
    	chosenfloor = -1;

    	if (isUpdate_ExreqUp.size() > 0) 
    		chosenfloor_up = GetFloortoMove_Ext(l.floor, isUpdate_ExreqUp, 0);

    	if (isUpdate_ExreqDown.size() > 0)
    		chosenfloor_down = GetFloortoMove_Ext(l.floor, isUpdate_ExreqDown, 0);

    	if ((chosenfloor_up != -1) && (chosenfloor_down != -1)) {
    		if (chosenfloor_up > chosenfloor_down) 
    			chosenfloor = chosenfloor_down;
    		else 
    			chosenfloor = chosenfloor_up;
    	}

    	else if (chosenfloor_up != -1) {
    		chosenfloor = chosenfloor_up;
    	}

    	else if (chosenfloor_down != -1) {
    		chosenfloor = chosenfloor_down;
    	}

    	else {
    		chosenfloor = -1;
    	}

    	if (chosenfloor != -1) {
    		if (l.floor > chosenfloor && l.floor > 0) {
				// go down
				LiftStatus[lift].direction = -1;
				LiftStatus[lift].floor = LiftStatus[lift].floor - 1;
			}
		
			if (l.floor < chosenfloor && l.floor < NumberofFloors - 1) {
				// go up
				LiftStatus[lift].direction = 1;
				LiftStatus[lift].floor = LiftStatus[lift].floor + 1;
			}
    	}
    }

    update_liftstatus.unlock();
}